﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Xml;
using System.IO;
using System.Text.RegularExpressions;
using Android.App;
using Cigno.SCS.Controls;

namespace Cigno.SCS.Control_App
{
    class Scene
    {
        XmlNode data;
        public string Name { get { return data.SelectSingleNode("@label").Value; } }
        public Scene(XmlNode x) { data = x; }
        public List<SceneAction> Actions {
            get
            {
                List<SceneAction> results = new List<SceneAction>();
                foreach (XmlNode x in data.SelectNodes("SCSAction"))
                {
                    results.Add(new SceneAction(x));
                }
                return results;
            }
        }
    }

    class SceneAction
    {
        public enum ActionTypeEnum
        {
            Command,
            WaitEvent
        }
        XmlNode data;
        public SceneAction(XmlNode x) { data = x; }
        public ActionTypeEnum ActionType { get { return string.IsNullOrEmpty(data.SelectSingleNode("@command").Value) ? ActionTypeEnum.WaitEvent : ActionTypeEnum.Command; } }
        public Int64 TimeOut { get { return Int64.Parse(data.SelectSingleNode("@timeout").Value); } }
        public string Command { get { return data.SelectSingleNode("@command").Value; } }
        public byte ControlID { get { return byte.Parse(data.SelectSingleNode("@controlID").Value, System.Globalization.NumberStyles.HexNumber); } }
    }
    class Control
    {
        XmlNode data;
        public byte ID { get { return byte.Parse(data.SelectSingleNode("@id").Value, System.Globalization.NumberStyles.HexNumber); } }
        public string Name { get { return data.SelectSingleNode("@label").Value; } }
        public SCSControlType Type
        {
            get
            {
                try
                {
                    return (SCSControlType)Enum.Parse(typeof(SCSControlType), data.SelectSingleNode("@type").Value);
                }
                catch (Exception e)
                {
                    return SCSControlType.Unknown;
                }
            }
        }
        public Control(XmlNode x) { data = x; }
    }
    class Page
    {
        XmlNode data;
        public byte Order { get { return byte.Parse(data.SelectSingleNode("@order").Value, System.Globalization.NumberStyles.HexNumber); } }
        public string Label { get { return data.SelectSingleNode("@label").Value; } }
        public Page(XmlNode x) { data = x; }
    }
    class Config
    {
        public XmlDocument Doc;
        public bool AutoDiscover {
            get {
                bool value = false;
                bool.TryParse(Doc.SelectSingleNode("SCSDeployment/SCSControls/@autoDiscover").Value, out value);
                return value;
            }
        }
        public string GatewayAddress { get { return Doc.SelectSingleNode("SCSDeployment/SCSDeploymentIdentifiers/SCSDeploymentIdentifier[@type=\"IP_Address\"][1]/@value").Value; } }
        public string GatewayMAC { get { return Doc.SelectSingleNode("SCSDeployment/SCSDeploymentIdentifiers/SCSDeploymentIdentifier[@type=\"MAC_Address\"][1]/@value").Value; } }
        public string Name { get { return Doc.SelectSingleNode("SCSDeployment/@name").Value; } }
        public string BroadcastName { get { return Doc.SelectSingleNode("SCSDeployment/SCSControls/@allControlsLabel").Value; } }
        private Dictionary<byte, Control> controls = null;
        public IDictionary<byte, Control> Controls
        {
            get
            {
                if (controls == null)
                {
                    controls = new Dictionary<byte, Control>();
                    foreach (XmlNode x in Doc.SelectNodes("SCSDeployment/SCSControls/SCSControl"))
                    {
                        Control control = new Control(x);
                        controls.Add(control.ID, control);
                    }
                }
                return controls;
            }
        }
        private Dictionary<string, Scene> scenes;
        public Dictionary<string, Scene> Scenes
        {
            get
            {
                if (scenes == null)
                {
                    scenes = new Dictionary<string, Scene>();
                    foreach (XmlNode x in Doc.SelectNodes("SCSDeployment/SCSScenes/SCSScene"))
                    {
                        scenes.Add(x.SelectSingleNode("@label").Value, new Scene(x));
                    }
                }
                return scenes;
            }
        }
        private Dictionary<string, Page> pages;
        public Dictionary<string,Page> Pages
        {
            get
            {
                if (pages == null)
                {
                    pages = new Dictionary<string, Page>();
                    foreach (XmlNode x in Doc.SelectNodes("SCSDeployment/SCSPage"))
                    {
                        pages.Add(x.SelectSingleNode("@label").Value, new Page(x));
                    }
                }
                return pages;
            }
        }
        /*public class ColorTheme {
            public Color OnOffButtonColor { get { return Android.Graphics.Color.SlateGray; } }
            public Color OnOffButtonTextColor { get { return Android.Graphics.Color.AntiqueWhite; } }
            public Color OnOffReadySliderColor { get { return Android.Graphics.Color.AliceBlue; } }
            public Color OnOffActiveSliderColor { get { return Android.Graphics.Color.BlueViolet; } }
            public Color OnOffSliderTextColor { get { return Android.Graphics.Color.Black; } }
            public Color MotorButtonColor { get { return Android.Graphics.Color.DarkGray; } }
            public Color MotorButtonTextColor { get { return Android.Graphics.Color.AntiqueWhite; } }
            public Color MotorReadySliderColor { get { return Android.Graphics.Color.DarkSeaGreen; } }
            public Color MotorActiveSliderColor { get { return Android.Graphics.Color.DarkOrange; } }
            public Color MotorSliderTextColor { get { return Android.Graphics.Color.GhostWhite; } }
            public Color ScenarioButtonColor { get { return Android.Graphics.Color.LightSlateGray; } }
            public Color ScenarioButtonTextColor { get { return Android.Graphics.Color.SaddleBrown; } }
            public Color ScenarioReadySliderColor { get { return Android.Graphics.Color.AliceBlue; } }
            public Color ScenarioActiveSliderColor { get { return Android.Graphics.Color.OrangeRed; } }
            public Color ScenarioSliderTextColor { get { return Android.Graphics.Color.Goldenrod; } }
            public Color BroadcastButtonColor { get { return Android.Graphics.Color.DarkGray; } }
            public Color BroadcastButtonTextColor { get { return Android.Graphics.Color.WhiteSmoke; } }
            public Color BroadcastReadySliderColor { get { return Android.Graphics.Color.SeaGreen; } }
            public Color BroadcastActiveSliderColor { get { return Android.Graphics.Color.OrangeRed; } }
            public Color BroadcastSliderTextColor { get { return Android.Graphics.Color.Goldenrod; } }

            public enum ThemedEntity
            {
                ButtonFace,
                ButtonText,
                SliderFace,
                SliderText
            }

            Color[SCSControlType,ThemedEntity,bool] {

            public GetColor(SCSControl Control, Status)
        };
        public ColorTheme Theme = new ColorTheme();*/

        public Config(XmlDocument x) { Doc = x; }
    }
    class Configs : Dictionary<String, Config>
    {
        string currentConfig = null;
        public Config Current {
            get
            {
                if (!string.IsNullOrEmpty(currentConfig) && this.ContainsKey(currentConfig)) {
                    return this[currentConfig];
                }
                else {
                    foreach (Config x in this.Values) {
                        return x;
                    }
                }
                return null;
            }
            set
            {
                if (ContainsValue(value))
                {
                    currentConfig = value.Name;
                }
            }
        }

        public void FindCurrent(SCSUDPGateway gw)
        {
            string ip = gw.EndPoint.Address.ToString();
            string mac = GetMACAddress( ip );
            foreach (Config c in this.Values)
            {
                if (string.Equals(c.GatewayMAC,mac,StringComparison.InvariantCultureIgnoreCase)) { currentConfig = c.Name; return; }
                if (string.Equals(c.GatewayAddress,ip)) { currentConfig = c.Name; return; }
            }
            currentConfig = null;
        }

        public string GetMACAddress(string ip) {
            try
            {
                string mac = null;
                System.IO.StreamReader ArpTable = new StreamReader(@"/proc/net/arp");
                while ( ArpTable.Peek()>=0 )
                {
                    string x = ArpTable.ReadLine();
                    if (x.StartsWith(ip+" "))
                    {
                        mac = Regex.Match(x, @"(?!\s)(([0-9,a-f,A-F]{2}(\:|\-)){5}[0-9,a-f,A-F]{2})(?!\t)").Value;
                        mac.Replace("-", ":");
                        break;
                    }
                }
                ArpTable.Close();
                return mac;
            } catch { }
            return null;
        }

        public Configs()
        {
            foreach (string x in Application.Context.Assets.List("Configuration"))
            {
                if (System.Text.RegularExpressions.Regex.Match(x, ".*\\.scsxml").Success)
                {
                    XmlDocument d = new XmlDocument();
                    d.Load(Application.Context.Assets.Open("Configuration/" + x));
                    this.Add(d);
                }
            }
        }
        public void Add(XmlDocument x)
        {
            Config xx = new Config(x);
            base.Add(xx.Name, xx);
        }
    }
}