﻿using System;
using Android.App;
using Android.Views;
using Android.Widget;
using Android.Graphics;
using Android.OS;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using Cigno.SCS.Controls;
using Android.Support.V4.Widget;

namespace Cigno.SCS.Control_App
{
    [Activity(Label = "SCS Remote", MainLauncher = true, Icon = "@drawable/icon", ScreenOrientation = Android.Content.PM.ScreenOrientation.Portrait)]
    public class SCSRemote : Activity
    {
        string GatewayAddress = null;
        SCSUDPGateway gw = null;
        public SCSControls Controls = null;
        Thread ControlThread = null;
        bool Requery = false;
        bool Paused = true;
        EventWaitHandle Resume = new EventWaitHandle(false, EventResetMode.AutoReset);
        private bool InitSCS()
        {
            // TODO implement configuration selector
            if (gw == null)
            {
                int Port = 52056; //fixed port
                IPEndPoint GatewayEndPoint = null;
                if (string.IsNullOrEmpty(GatewayAddress)) { GatewayAddress = configs.Current.GatewayAddress; }
                if (string.IsNullOrEmpty(GatewayAddress) || GatewayAddress == "0.0.0.0")
                {
                    List<IPEndPoint> EPs = new List<IPEndPoint>();
                    SCSUDPGateway.AutoDiscover(EPs);
                    if (null != EPs && 0 < EPs.Count)
                    {
                        GatewayEndPoint = EPs[0];
                    }
                }
                else
                {
                    GatewayEndPoint = SCSUDPGateway.GetGatewayEndPoint(GatewayAddress, Port, 500);
                }
                if (GatewayEndPoint != null)
                {
                    gw = new SCSUDPGateway(GatewayEndPoint);
                    configs.FindCurrent(gw);

                    Controls = new SCSControls();
                    if (configs.Current.Controls.Count == 0 || configs.Current.AutoDiscover) { Controls.Discover(gw); }

                    foreach (Control control in configs.Current.Controls.Values)
                    {
                        SCSControl scscontrol = Controls.Item(control.ID);
                        if (scscontrol != null)
                        { // existing control: rename/repurpose
                            Controls.Rename(control.ID, control.Name);
                            if (control.Type != SCSControlType.Unknown) { Controls.Repurpose(control.ID, control.Type); }
                        }
                        else
                        { // non-existing control: add new
                            Controls.Add(new SCSControl(gw, control.Name, control.ID, control.Type == SCSControlType.Unknown ? SCSControlType.OnOff : control.Type));
                        }
                    }
                    Controls.Add(new SCSControl(gw, configs.Current.BroadcastName, 0xb1, SCSControlType.Broadcast));

                    foreach (SCSControl x in Controls.Values)
                    {
                        x.OnStatusChanged += StatusChangedHandler;
                    }
                    return true;
                } else {
                    return false;
                }
            }
            else
            {
                return true;
            }
        }
        /*private void ResumeSCS()
        {
            Paused = false;
        }
        private void PauseSCS()
        {
            Paused = true;
        }*/
        private void ReleaseSCS()
        {
            if (gw != null) gw.StopReceiveLoop();
            foreach (SCSControl x in Controls.Values)
            {
                x.OnStatusChanged -= StatusChangedHandler;
            }
            Controls = null;
            gw = null;
        }
        Configs configs = new Configs();
        private DateTime LastBackPressed = DateTime.MinValue;
        private void MainLoop()
        {
            // STEP 1: load config
            //SetStatus("Loading saved config...", 3000);

            // STEP 2: Loop
            do
            {
                if (gw==null && Requery && !Paused)
                {
                    // STEP 2.1: Inform user
                    status.Show("Initializing SCS gateway", 3000);

                    // STEP 2.2: Initialize SCS
                    if (InitSCS())
                    {
                        status.Show("SCS gateway initialization successful. Initializing UI", 3000);
                    }
                    else
                    {
                        status.Show("SCS gateway initialization failed", 30000);
                    };

                    // STEP 2.3: Update UI
                    InitializeUI();
                }

                // STEP 2.4 Update status
                //UpdateStatus();
                // STEP 2.5 Stop/resume gateway receive loop
                if (Paused && gw != null && gw.IsReceiving) {
                    status.Show("Going to sleep", 3000);
                    gw.StopReceiveLoop();
                }
                if (!Paused && gw != null && !gw.IsReceiving) {
                    status.Show("Resuming...", 3000);
                    gw.StartReceiveLoop();
                }
                // STEP 2.6 run commands
                if (!Paused && gw != null)
                {
                    if (Requery)
                    {
                        status.Show("Refreshing", 3000);
                        //Controls[configs.Current.BroadcastName].GetStatus();
                        Controls.Broadcast.GetStatus();
                        Requery = false;
                    }
                    /*for (int i = 0; i < 3 && Commands.Count > 0; i++)
                    {
                        Command x = Commands.Dequeue();
                        //SetStatus(string.Format("{0}:{1}", x.Control.Name, x.Action), 3000);
                        //x.Control.SetStatus(x.Action);
                        x.Fire();
                    }*/
                    Commands.Fire(3);
                    //Thread.Sleep(100);
                    Thread.Yield();
                    if (Commands.Count == 0)
                    {
                        Commands.ItemsPending.WaitOne(3000);
                    }
                }
                else
                {
                    //Thread.Sleep(100);
                    Thread.Yield();
                    Resume.WaitOne(1000);
                }
            } while (true);
        }
        #region "Activity lifecycle"

        Status status = null;

        protected override void OnCreate(Bundle bundle)
        {
            RequestWindowFeature(WindowFeatures.NoTitle);
            Window.AddFlags(WindowManagerFlags.Fullscreen);
            Window.AddFlags(WindowManagerFlags.KeepScreenOn);
            Window.AddFlags(WindowManagerFlags.DismissKeyguard);
            Window.AddFlags(WindowManagerFlags.ShowWhenLocked);
            base.OnCreate(bundle);
            status = new Status(this);

            /*if (bundle != null)
            {
                configs.Current = configs[bundle.GetString(cLastActiveConfig)];
                GatewayAddress = bundle.GetString(cGatewayEndPoint);
            }*/
            string x = Resources.GetLayout(Resource.Layout.SCSRemote).ReadOuterXml();
            SetContentView(Resource.Layout.SCSRemote);

            //if (ControlThread != null && ControlThread.IsAlive) { return; }
            ControlThread = new Thread(MainLoop);
            ControlThread.Start();
        }
        protected override void OnResume()
        {
            base.OnResume();
            Requery = true;
            Paused = false;
            Resume.Set();
            Commands.Clear();
        }
        protected override void OnPause()
        {
            base.OnPause();
            Paused = true;
        }
        const string cLastActiveConfig = "LAST_ACTIVE_CONFIG";
        const string cGatewayEndPoint = "GATEWAY_ENDPOINT";
        /*protected override void OnSaveInstanceState(Bundle outState)
        {
            outState.PutString(cLastActiveConfig, configs.Current.Name);
            outState.PutString(cGatewayEndPoint, gw!=null?gw.EndPoint.Address.ToString():"0.0.0.0");
            base.OnSaveInstanceState(outState);
        }*/
        protected override void OnDestroy()
        {
            if (ControlThread != null && ControlThread.ThreadState != System.Threading.ThreadState.Aborted)
            {
                ControlThread.Abort();
            }
            ControlThread = null;
            status.Suspend();
            base.OnDestroy();
        }
        protected override void OnStop()
        {
            status.Suspend();
            base.OnStop();
        }
        public override void OnBackPressed()
        {
            if (LastBackPressed == null || (DateTime.Now - LastBackPressed).Milliseconds > 300)
            {
                status.Show("Press back again to close", 500);
                LastBackPressed = DateTime.Now;
            }
            else
            {
                base.OnBackPressed();
            }
        }
        /*public override void OnAttachedToWindow()
        {
            Window.SetType(WindowManagerTypes.KeyguardDialog);
            base.OnAttachedToWindow();
        }*/
        #endregion "Activity lifecycle"
        #region "UI"
        private void InitializeUI()
        {
            LinearLayout container = FindViewById<LinearLayout>(Resource.Id.MainContainer);
            /*this.RunOnUiThread(() => {
                LinearLayout.LayoutParams p = (LinearLayout.LayoutParams)(container.LayoutParameters);
                p.SetMargins(50, 10, 50, 10);
            });*/
            if (Controls != null)
            {
                foreach (Scene x in configs.Current.Scenes.Values)
                {
                    this.RunOnUiThread(() =>
                    {
                        /*ToggleButton btn = new ToggleButton(container.Context)
                        {
                            Text = x.Name,
                            TextOn = x.Name,
                            TextOff = x.Name,
                            Tag = x.Name,
                            Checked = false,
                            HapticFeedbackEnabled = true
                        };
                        btn.Click += SceneClickHandler;*/
                        SwipeButton btn = new SwipeButton(container.Context)
                        {
                            Tag = x.Name,
                            ButtonName = x.Name,
                            ButtonStatusText = "",
                            HapticFeedbackEnabled = true,
                            IndicatorColor = Color.SeaGreen,
                            ButtonColor = Color.Pink
                        };
                        btn.ActiveChange += SceneClickHandler;

                        container.AddView(btn);
                        container.AddView(new Android.Widget.Space(container.Context));
                        container.AddView(new Android.Widget.Space(container.Context), new ViewGroup.LayoutParams(-2, 8));
                    });
                }

                foreach (Control x in configs.Current.Controls.Values)
                {
                    //if (x.Type == SCSControlType.Dimmer || x.Type== SCSControlType.OnOff)
                    {
                        this.RunOnUiThread(() =>
                        {
                            /*ToggleButton btn = new ToggleButton(container.Context)
                            {
                                Text = x.Name + ": Unknown",
                                TextOn = x.Name + ": Unknown",
                                TextOff = x.Name + ": Unknown",
                                Tag = x.Name,
                                Checked = false,
                                HapticFeedbackEnabled = true
                            };
                            btn.Click += ControlClickHandler;*/
                            SwipeButton btn = new SwipeButton(container.Context)
                            {
                                ButtonName = x.Name,
                                ButtonStatusText = "Unknown",
                                Tag = x.Name,
                                HapticFeedbackEnabled = true,
                                IndicatorColor = Color.DarkSlateBlue,
                                Actions =  Controls.Item(x.ID).Commands,
                                NextAction = Controls.Item(x.ID).DefaultCommand
                            };
                            btn.ActiveChange += ControlClickHandler;
                            container.AddView(btn);
                            container.AddView(new Android.Widget.Space(container.Context),new ViewGroup.LayoutParams(-2,8));
                        });
                    }
                }
                /* CREATE BUTTONS FOR AUTODISCOVERED SCS CONTROLS
                 * 
                 * foreach (SCSControl x in Controls.Values)
                {
                    if (x.ControlType == SCSControlType.Dimmer || x.ControlType == SCSControlType.OnOff)
                    {
                        this.RunOnUiThread(() =>
                        {
                            ToggleButton btn = new ToggleButton(container.Context)
                            {
                                Text = x.Name + ": Unknown",
                                TextOn = x.Name + ": Unknown",
                                TextOff = x.Name + ": Unknown",
                                Tag = x.Name,
                                Checked = false
                            };
                            btn.Click += ClickHandler;
                            container.AddView(btn);
                        });
                    }
                }*/
                SwipeRefreshLayout refresher = FindViewById<SwipeRefreshLayout>(Resource.Id.MainLayout);
                refresher.Refresh += RefreshHandler;
            }
        }

        protected class Status
        {
            private SCSRemote activity;
            private Toast StatusToast;
            private System.Timers.Timer StatusTimer = null;
            private DateTime StatusExpires = DateTime.MaxValue;

            public Status(SCSRemote activity)
            {
                this.activity = activity;
                StatusTimer = new System.Timers.Timer() { Interval = 100, Enabled = false };
                StatusTimer.Elapsed += StatusTimerCallBack;
                StatusExpires = DateTime.MaxValue;
            }
            private void StatusTimerCallBack(object sender, System.Timers.ElapsedEventArgs e)
            {
                if (DateTime.Now >= StatusExpires)
                {
                    Clear();
                }
            }
            public void Show(string Value, int Duration)
            {
                if (StatusToast != null) Clear();

                activity.RunOnUiThread(() =>
                {
                    StatusToast = Toast.MakeText(activity.ApplicationContext, Value, ToastLength.Long);
                    StatusToast.Show();
                });
                StatusExpires = DateTime.Now.AddMilliseconds(Duration);
                StatusTimer.Enabled = true;
                StatusTimer.Start();
            }

            public void Suspend()
            {
                Clear();
            }
            private void Clear()
            {
                StatusTimer.Stop();
                StatusTimer.Enabled = false;
                activity.RunOnUiThread(() => { StatusToast.Cancel(); });
            }
        }
        #endregion "UI"
        #region "UI callback handlers"
        private void StatusChangedHandler(object sender, string e)
        {
            LinearLayout container = FindViewById<LinearLayout>(Resource.Id.MainContainer);
            /*ToggleButton btn = (ToggleButton)container.FindViewWithTag(((Java.Lang.Object)((SCSControl)sender).Name));
            if (btn != null)
            {
                this.RunOnUiThread(() =>
                {
                    btn.Checked = e != "OFF";
                    btn.Text = ((SCSControl)sender).Name + ": " + e;
                });
            }*/
            SwipeButton btn = (SwipeButton)container.FindViewWithTag(((Java.Lang.Object)((SCSControl)sender).Name));
            if (btn!=null)
            {
                this.RunOnUiThread(() => 
                {
                    btn.ButtonStatusText = e;
                    btn.IndicatorColor = e != "OFF" ? Color.DarkSlateBlue: Color.BlueViolet;
                    //btn.Reset();
                });
            }
        }

        private void RefreshHandler(object sender,EventArgs e)
        {
            if (gw!=null)
            {
                status.Show("Refreshing", 3000);
                Controls.Broadcast.GetStatus();
                //Commands.Enqueue(new Command(this, Controls.Broadcast.Name, "_REFRESH"));
            } else
            {
                Requery = true;
            }
            ((SwipeRefreshLayout)sender).Refreshing = false;
        }

        private void ControlClickHandler(object sender, EventArgs e)
        {
            //SCSControl x = Controls[(string)(((View)sender).Tag)];
            //x.SetStatus("TOGGLE");
            //x.SetStatus(x.DefaultCommand);
            SwipeButton btn = (SwipeButton)sender;
            SCSControl ctrl = Controls[(string)btn.Tag];
            string cmd = btn.NextAction;
            if (cmd!=ctrl.DefaultCommand)
            {
                btn.NextAction = ctrl.DefaultCommand;
            }
            Commands.Enqueue(new Command(this, ctrl.Name, cmd));
            //Commands.Enqueue(new Command(this, (string)(((View)sender).Tag), null));
            //((ToggleButton) sender).Text = x.Name + ": " + x.LastReadStatus;
        }

        private void SceneClickHandler(object sender, EventArgs e)
        {
            if (configs.Current.Scenes.ContainsKey((string)(((View)sender).Tag)))
            {
                foreach (SceneAction x in configs.Current.Scenes[(string)(((View)sender).Tag)].Actions)
                { 
                    if (x.ActionType==SceneAction.ActionTypeEnum.Command)
                    {
                        Commands.Enqueue(new Command(this,Controls.Item(x.ControlID).Name,x.Command));
                    }
                    else
                    {
                        //TODO: handle events
                    }
                }
            }
        }
        #endregion "UI callback handlers"
        #region "Queuing"
        protected class Command
        {
            //public SCSControl Control { get; set; }
            public string Control { get; set; }
            public string Action { get; set; }
            public bool IsUI { get; set; }
            public SCSRemote Parent { get; set; }
            //public Command(SCSControl Control,string Action) { this.Control = Control;this.Action = Action; }
            public Command(SCSRemote Parent, string Control, string Action) { this.Parent = Parent; this.Control = Control; this.Action = Action; }
            public Command(string UIControl, string Action) { this.Control = UIControl; IsUI = true; this.Action = Action; }
            public event EventHandler Fired;
            public void Fire()
            {
                if (!IsUI) {
                    if (Parent.Controls.TryGetValue(Control, out SCSControl x)) {
                        if (string.IsNullOrEmpty(Action)) { Action = x.DefaultCommand; }
                        Parent.status.Show(string.Format("{0}:{1}", Control, Action), x.ControlType == SCSControlType.Broadcast ? 500 : 3000);
                        Fired?.Invoke(this, EventArgs.Empty);
                        x.SetStatus(Action);
                    }
                } else {
                    switch (Action) { 
                        case "REQUERY":
                            Parent.Controls.Broadcast.GetStatus();
                            break;
                        default: 
                            Parent.RunOnUiThread(() => {
                                LinearLayout container = Parent.FindViewById<LinearLayout>(Resource.Id.MainContainer);
                                SwipeButton btn = (SwipeButton)container.FindViewWithTag((Java.Lang.Object)Control);
                                //if (btn!=null) { btn.Reset(); }
                            });
                            break;
                    }
                }
            }
        }
        protected class CommandQueue : Queue<Command>
        {
            public EventWaitHandle ItemsPending = new EventWaitHandle(false, EventResetMode.ManualReset);
            public new void Enqueue(Command item)
            {
                base.Enqueue(item);
                ItemsPending.Set();
            }
            /*private new Command Dequeue()
            {
                if (Count == 0) ItemsPending.Close();
                return base.Dequeue();
            }*/
            public void Fire(int count)
            {
                for (int i = 0; i < 3 && Count > 0; i++)
                {
                    Command x = Dequeue();
                    x.Fire();
                }
            }
        }
        //Queue<Command> Commands = new Queue<Command>();
        CommandQueue Commands = new CommandQueue();
        #endregion "Queuing"
    }
}

