﻿using System;
using Android.Content;
using Android.Views;
using Android.Widget;
using Android.Util;
using Android.Animation;
using Android.Content.Res;
using Android.Graphics;
using Android.Graphics.Drawables;
using Android.Support.V4.Content;
using Android.Views.Animations;
using Android.OS;
using Android.App;

namespace Cigno.SCS.Control_App
{
    class SwipeButton : RelativeLayout, View.IOnTouchListener, View.IOnLongClickListener, IDialogInterfaceOnClickListener, IDialogInterfaceOnCancelListener,IDialogInterfaceOnDismissListener
    {
        #region Declarations
        // Declarations: animation
        private Handler DelayedResetHandler = new Handler();

        // Declarations: Spinner
        private AlertDialog.Builder spinner;
        private ArrayAdapter<string> ActionsAdapter;
        private bool spinnerShowing = false;
        private bool longClickDetection = false;
        private Handler LongClickDetectorHandler = new Handler();

        // Declarations: UI elements and properties
        private int collapsedWidth;
        private int collapsedHeight;
        private LinearLayout layer;
        private bool trailEnabled = false;
        private TextView centerText;
        private TextView buttonText;
        private ViewGroup background;
        public override Drawable Background { set => background.Background = value; }
        public Drawable SlidingButtonBackground { set => background.Background = value; }
        private Drawable disabledDrawable;
        private Drawable enabledDrawable;
        private ImageView swipeButtonImage;
        private RelativeLayout swipeButtonInner;
        public Drawable ButtonBackground { set { if (value != null) { swipeButtonInner.Background = value; } } get { return swipeButtonInner.Background; } }

        // Declarations: MISC
        //private static readonly int ENABLED = 0;
        //private static readonly int DISABLED = 1;
        #endregion

        #region Properties
        private string name = "";
        private string status = "";
        public string ButtonName
        {
            set
            {
                name = value;
                centerText.SetText(name + (string.IsNullOrEmpty(status) ? "" : " : " + status), TextView.BufferType.Normal);
            }
            private get
            {
                return name;
            }
        }
        public string ButtonStatusText
        {
            set
            {
                status = value;
                //buttonText.SetText(status, TextView.BufferType.Normal);
                centerText.SetText(name + (string.IsNullOrEmpty(status) ? "" : " : " + status), TextView.BufferType.Normal);
            }
            private get
            {
                return status;
            }
        }
        public string NextAction
        {
            set
            {
                buttonText.SetText(value, TextView.BufferType.Normal);
            }
            get
            {
                return buttonText.Text;
            }
        }
        public Color ButtonColor
        {
            set
            {
                //this.SetBackgroundColor(value);
                //this.Background.SetTint(value.ToArgb());
                //this.background.SetBackgroundColor(value);
                //this.Background.SetTint(value.ToArgb());
                //background.GetChildAt(0).Background.SetTint(value.ToArgb());
            }
        }
        public Color IndicatorColor
        {
            set
            {
                swipeButtonInner.Background.SetTint(value.ToArgb());
            }
        }

        public System.Collections.Generic.IEnumerable<string> Actions
        {
            set
            {
                ActionsAdapter.Clear();
                NextAction = string.Empty;
                foreach (string s in value) {
                    ActionsAdapter.Add(s);
                }
            }
        }
        public enum ButtonState
        {
            InitialPosition,
            Moving,
            Expanded
        };

        public ButtonState UIStatus { get; private set; } = ButtonState.InitialPosition;
        #endregion

        #region Base UI
        // Constructor
        public SwipeButton(Context context):base(context)
        { Init(context, null, -1, -1); }
        public SwipeButton(Context context, IAttributeSet attrs) : base(context, attrs)
        { Init(context, attrs, -1, -1); }
        public SwipeButton(Context context, IAttributeSet attrs, int defStyleAttr) : base(context, attrs, defStyleAttr)
        { Init(context, attrs, defStyleAttr, -1); }
        public SwipeButton(Context context, IAttributeSet attrs, int defStyleAttr, int defStyleRes) : base(context, attrs, defStyleAttr, defStyleRes)
        { Init(context, attrs, defStyleAttr, defStyleRes); }

        public void SetDisabledDrawable(Drawable drawable)
        {
            disabledDrawable = drawable;

            if (ButtonState.InitialPosition==UIStatus)
            {
                swipeButtonImage.SetImageDrawable(drawable);
            }
        }
        public void SetEnabledDrawable(Drawable drawable)
        {
            enabledDrawable = drawable;

            if (ButtonState.InitialPosition==UIStatus)
            {
                swipeButtonImage.SetImageDrawable(drawable);
            }
        }

        //public event EventHandler StateChange;
        public event EventHandler ActiveChange;

        public void SetTextPadding(int left, int top, int right, int bottom)
        {
            centerText.SetPadding(left, top, right, bottom);
        }

        public void SetSwipeButtonPadding(int left, int top, int right, int bottom)
        {
            swipeButtonInner.SetPadding(left, top, right, bottom);
        }

        public bool HasActivationState { set; get; }
        public int DpToPx(int dp)
        {
            return (int)Math.Round(dp * (Context.Resources.DisplayMetrics.Xdpi / (int)DisplayMetricsDensity.Default));
        }

        public float PxToSp(float px,Context context)
        {
            return px / context.Resources.DisplayMetrics.ScaledDensity;
        }

        private void Init(Context context, IAttributeSet attrs, int defStyleAttr, int defStyleRes)
        {
            HasActivationState = true;
            this.LongClickable = true;

            // Background
            background = new RelativeLayout(context);
            LayoutParams layoutParamsView = new LayoutParams(ViewGroup.LayoutParams.MatchParent, ViewGroup.LayoutParams.WrapContent);
            layoutParamsView.AddRule(LayoutRules.CenterInParent);
            AddView(background, layoutParamsView);

            // Center Text
            centerText = new TextView(context) {
                Gravity = GravityFlags.Center
            };
            LayoutParams layoutParams = new LayoutParams(ViewGroup.LayoutParams.WrapContent, ViewGroup.LayoutParams.WrapContent);
            layoutParams.AddRule(LayoutRules.CenterInParent);
            background.AddView(centerText, layoutParams);

            // Swipe Button - Container
            swipeButtonInner = new RelativeLayout(context);
            layoutParamsView = new LayoutParams(ViewGroup.LayoutParams.MatchParent, ViewGroup.LayoutParams.WrapContent);
            layoutParamsView.AddRule(LayoutRules.CenterInParent);

            // Swipe Button - Image
            swipeButtonImage = new ImageView(context) {
                LayoutParameters = new ViewGroup.LayoutParams(-2, -2)
            };
            swipeButtonInner.AddView(swipeButtonImage);

            // Swipe Button - Text
            buttonText = new TextView(context) {
                Gravity = GravityFlags.Center
            };
            swipeButtonInner.AddView(buttonText, new LayoutParams(ViewGroup.LayoutParams.WrapContent, ViewGroup.LayoutParams.WrapContent));

            /*if (attrs != null && defStyleAttr == -1 && defStyleRes == -1) // this portion of code is not maintained
            {
                TypedArray typedArray = context.ObtainStyledAttributes(attrs, Resource.Styleable.SwipeButton, defStyleAttr, defStyleRes);
                collapsedWidth = (int)typedArray.GetDimension(Resource.Styleable.SwipeButton_button_image_width, ViewGroup.LayoutParams.WrapContent);
                collapsedHeight = (int)typedArray.GetDimension(Resource.Styleable.SwipeButton_button_image_height, ViewGroup.LayoutParams.WrapContent);
                trailEnabled = typedArray.GetBoolean(Resource.Styleable.SwipeButton_button_trail_enabled, false);
                Drawable trailingDrawable = typedArray.GetDrawable(Resource.Styleable.SwipeButton_button_trail_drawable);
                Drawable backgroundDrawable = typedArray.GetDrawable(Resource.Styleable.SwipeButton_inner_text_background);

                if (backgroundDrawable != null)
                {
                    background.Background=backgroundDrawable;
                } else {
                    background.Background=ContextCompat.GetDrawable(context, Resource.Drawable.shape_rounded);
                }

                if (trailEnabled)
                {
                    layer = new LinearLayout(context);

                    if (trailingDrawable != null)
                    {
                        layer.Background=trailingDrawable;
                    } else {
                        layer.Background=typedArray.GetDrawable(Resource.Styleable.SwipeButton_button_background);
                    }

                    layer.SetGravity(GravityFlags.Start);
                    layer.Visibility=ViewStates.Gone;
                    background.AddView(layer, layoutParamsView);
                }
                //centerText.SetText(Resource.Styleable.SwipeButton_inner_text);
                centerText.SetText(typedArray.GetString(Resource.Styleable.SwipeButton_inner_text), TextView.BufferType.Normal);
                centerText.SetTextColor(typedArray.GetColor(Resource.Styleable.SwipeButton_inner_text_color, Color.White));

                float textSize = PxToSp( typedArray.GetDimension(Resource.Styleable.SwipeButton_inner_text_size, 0), context);

                centerText.TextSize = (textSize != 0) ? textSize:12;

                disabledDrawable = typedArray.GetDrawable(Resource.Styleable.SwipeButton_button_image_disabled);
                enabledDrawable = typedArray.GetDrawable(Resource.Styleable.SwipeButton_button_image_enabled);
                float innerTextLeftPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_inner_text_left_padding, 0);
                float innerTextTopPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_inner_text_top_padding, 0);
                float innerTextRightPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_inner_text_right_padding, 0);
                float innerTextBottomPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_inner_text_bottom_padding, 0);

                int initialState = typedArray.GetInt(Resource.Styleable.SwipeButton_initial_state, DISABLED);

                if (initialState == ENABLED)
                {
                    LayoutParams layoutParamsButton = new LayoutParams( ViewGroup.LayoutParams.MatchParent, ViewGroup.LayoutParams.WrapContent );
                    layoutParamsButton.AddRule(LayoutRules.AlignParentLeft);
                    layoutParamsButton.AddRule(LayoutRules.CenterVertical);
                    swipeButtonImage.SetImageDrawable(enabledDrawable);
                    AddView(swipeButtonInner, layoutParamsButton);
                    UIStatus = ButtonState.Expanded;
                    //Active = true;
                } else {
                    LayoutParams layoutParamsButton = new LayoutParams(collapsedWidth, collapsedHeight);
                    layoutParamsButton.AddRule(LayoutRules.AlignParentLeft);
                    layoutParamsButton.AddRule(LayoutRules.CenterVertical);
                    swipeButtonImage.SetImageDrawable(disabledDrawable);
                    AddView(swipeButtonInner, layoutParamsButton);
                    UIStatus = ButtonState.InitialPosition;
                    //Active = false;
                }

                centerText.SetPadding((int)innerTextLeftPadding, (int)innerTextTopPadding, (int)innerTextRightPadding, (int)innerTextBottomPadding);

                Drawable buttonBackground = typedArray.GetDrawable(Resource.Styleable.SwipeButton_button_background);

                swipeButtonInner.Background = (buttonBackground != null) ? buttonBackground:ContextCompat.GetDrawable(context, Resource.Drawable.shape_button);

                float buttonLeftPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_button_left_padding, 0);
                float buttonTopPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_button_top_padding, 0);
                float buttonRightPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_button_right_padding, 0);
                float buttonBottomPadding = typedArray.GetDimension(Resource.Styleable.SwipeButton_button_bottom_padding, 0);

                swipeButtonInner.SetPadding((int)buttonLeftPadding, (int)buttonTopPadding, (int)buttonRightPadding, (int)buttonBottomPadding);

                HasActivationState = typedArray.GetBoolean(Resource.Styleable.SwipeButton_has_activate_state, true);

                typedArray.Recycle();
            }
            else // attr is empty: expected default*/
            {
                collapsedWidth = 200;
                collapsedHeight = 100;
                trailEnabled = false;
                background.Background = ContextCompat.GetDrawable(context, Resource.Drawable.shape_rounded);
                background.SetMinimumHeight(collapsedHeight);
                float backgroundLeftPadding = 20;
                float backgroundTopPadding = 20;
                float backgroundRightPadding = 20;
                float backgroundBottomPadding = 20;
                background.SetPadding((int)backgroundLeftPadding, (int)backgroundTopPadding, (int)backgroundRightPadding, (int)backgroundBottomPadding);

                centerText.SetTextColor(Color.White);
                centerText.TextSize = 12;
                float innerTextLeftPadding = 20;
                float innerTextTopPadding = 20;
                float innerTextRightPadding = 20;
                float innerTextBottomPadding = 20;
                centerText.SetPadding((int)innerTextLeftPadding, (int)innerTextTopPadding, (int)innerTextRightPadding, (int)innerTextBottomPadding);

                buttonText.SetTextColor(Color.Black);
                buttonText.TextSize = 12;
                buttonText.SetPadding((int)innerTextLeftPadding, (int)innerTextTopPadding, (int)innerTextRightPadding, (int)innerTextBottomPadding);

                //int initialState = DISABLED;

                LayoutParams layoutParamsButton = new LayoutParams(collapsedWidth, collapsedHeight);
                layoutParamsButton.AddRule(LayoutRules.AlignParentLeft);
                layoutParamsButton.AddRule(LayoutRules.CenterVertical);
                swipeButtonImage.SetImageDrawable(disabledDrawable);
                AddView(swipeButtonInner, layoutParamsButton);
                //Active = false;
                UIStatus = ButtonState.InitialPosition;

                swipeButtonInner.Background = ContextCompat.GetDrawable(context, Resource.Drawable.shape_button);
                swipeButtonInner.SetMinimumHeight(collapsedHeight);
                swipeButtonInner.SetMinimumWidth(collapsedWidth);

                HasActivationState = true;
            }

            // Initialize Spinner
            ActionsAdapter = new ArrayAdapter<string>(context, Android.Resource.Layout.SimpleListItem1);
            spinner = new AlertDialog.Builder(context);
            spinner.SetOnCancelListener(this);
            spinner.SetOnDismissListener(this);

            this.SetOnTouchListener(this);
            //this.SetOnLongClickListener(this);
        }
        #endregion

        #region Animations
        private void ExpandButton()
        {
            ValueAnimator positionAnimator = ValueAnimator.OfFloat( swipeButtonInner.GetX(), 0 );
            positionAnimator.Update += (v, e) => {
                float x = (float)positionAnimator.AnimatedValue;
                swipeButtonInner.SetX(x);
            };
            ValueAnimator widthAnimator = ValueAnimator.OfInt( swipeButtonInner.Width, Width );
            widthAnimator.Update += (v, e) => {
                ViewGroup.LayoutParams p = swipeButtonInner.LayoutParameters;
                if (p!=null)
                {
                    p.Width = ((int)widthAnimator.AnimatedValue);
                    swipeButtonInner.LayoutParameters = p;
                }
            };

            AnimatorSet animatorSet = new AnimatorSet();
            animatorSet.AnimationEnd += (v, e) => {
                swipeButtonImage.SetImageDrawable(enabledDrawable);
            };

            animatorSet.PlayTogether(positionAnimator, widthAnimator);
            UIStatus = ButtonState.Expanded;
            animatorSet.Start();
        }

        private void MoveButtonBack()
        {
            ValueAnimator positionAnimator = ValueAnimator.OfFloat(swipeButtonInner.GetX(), 0);
            positionAnimator.SetInterpolator(new AccelerateDecelerateInterpolator());
            positionAnimator.Update += (object sender, ValueAnimator.AnimatorUpdateEventArgs e) => {
                float x = (float)positionAnimator.AnimatedValue;
                swipeButtonInner.SetX(x);
                SetTrailingEffect();
            };
            positionAnimator.AnimationEnd += (v,e) => {
                if (layer != null) {
                    layer.Visibility=ViewStates.Gone;
                }
            };

            ObjectAnimator objectAnimator = ObjectAnimator.OfFloat(centerText, "alpha", 1);

            positionAnimator.SetDuration(200);

            AnimatorSet animatorSet = new AnimatorSet();
            animatorSet.PlayTogether(objectAnimator, positionAnimator);
            animatorSet.Start();
            UIStatus = ButtonState.InitialPosition;
        }

        private void CollapseButton()
        {
            int finalWidth;

            if (collapsedWidth == ViewGroup.LayoutParams.WrapContent) {
                finalWidth = swipeButtonInner.Height;
            } else {
                finalWidth = collapsedWidth;
            }
            ValueAnimator widthAnimator = ValueAnimator.OfInt(swipeButtonInner.Width, finalWidth);
            widthAnimator.Update += (v, e) => {
                ViewGroup.LayoutParams p = swipeButtonInner.LayoutParameters;
                p.Width = (int)widthAnimator.AnimatedValue;
                swipeButtonInner.LayoutParameters = p;
                SetTrailingEffect();
            };
            widthAnimator.AnimationEnd += (v, e) => {
                swipeButtonImage.SetImageDrawable(disabledDrawable);
                if (layer != null) {
                    layer.Visibility = ViewStates.Gone;
                }
            };
            ObjectAnimator objectAnimator = ObjectAnimator.OfFloat(centerText, "alpha", 1);

            AnimatorSet animatorSet = new AnimatorSet();
            animatorSet.PlayTogether(objectAnimator, widthAnimator);
            animatorSet.Start();
            UIStatus = ButtonState.InitialPosition;
        }

        private void SetTrailingEffect()
        {
            if (trailEnabled) {
                layer.Visibility=ViewStates.Visible;
                layer.LayoutParameters=new RelativeLayout.LayoutParams((int)(swipeButtonInner.GetX() + swipeButtonInner.Width / 3), centerText.Height);
            }
        }
        #endregion

        #region Button Status
        private void InvokeDelayedReset()
        {
            DelayedResetHandler.PostDelayed(Reset, ButtonState.Expanded == UIStatus ? 400 : 200);
        }

        private void CancelDelayedReset()
        {
            DelayedResetHandler.RemoveCallbacks(Reset);
        }

        private void BeginSwipe() // TODO: use this to block parent scrolling while swipe button in motion
        {
            UIStatus = ButtonState.Moving;
            this.Parent.RequestDisallowInterceptTouchEvent(true);
        }

        private void EndSwipe()
        {
            this.Parent.RequestDisallowInterceptTouchEvent(false);
        }

        private void Reset()
        {
            switch (UIStatus)
            {
                case ButtonState.Moving:
                    MoveButtonBack();
                    break;
                case ButtonState.Expanded:
                    if (HasActivationState)
                    {
                        CollapseButton();
                    } else
                    {
                        MoveButtonBack();
                    }
                    break;
            }
            CancelDelayedReset();
        }

        private void SetActive()
        {
            if (ButtonState.Moving==UIStatus)
            {
                if (HasActivationState)
                {
                    ExpandButton();
                }
                ActiveChange?.Invoke(this, EventArgs.Empty);
                InvokeDelayedReset();
            }
        }

        /*public void ToggleState()
        {
            //CanMove = false;
            switch (UIStatus)
            {
                case ButtonState.Moving:
                    ActiveChange(this, EventArgs.Empty);
                    ExpandButton();
                    DelayedActions.Start();
                    break;
                case ButtonState.Expanded:
                    CollapseButton();
                    break;
            }
        }*/
        #endregion

        #region Spinner
        void IDialogInterfaceOnClickListener.OnClick(IDialogInterface dialog, int which)
        {
            spinnerShowing = false;
            NextAction = ActionsAdapter.GetItem(which);
            dialog.Dismiss();
        }
        public void OnCancel(IDialogInterface dialog)
        {
            spinnerShowing = false;
            //dialog.Dismiss();
        }

        void IDialogInterfaceOnDismissListener.OnDismiss(IDialogInterface dialog)
        {
            spinnerShowing = false;
            //dialog.Dismiss();
        }
        public bool OnLongClick(View v)
        {
            Reset();
            if (!spinnerShowing && ActionsAdapter!=null && ActionsAdapter.Count>0)
            {
                spinnerShowing = true;
                spinner.SetSingleChoiceItems(ActionsAdapter, -1, this);
                spinner.Show();
            }
            return true;
        }
        #endregion

        #region Touch listeners

        private void StartLongClickDetection()
        {
            System.Diagnostics.Debug.WriteLine("Starting long click detection");
            longClickDetection = true;
            LongClickDetectorHandler.PostDelayed(LongClickDetector, 1000);
        }

        private void LongClickDetector()
        {
            System.Diagnostics.Debug.WriteLine("{0}: {1} - Long click callback called", DateTime.Now, name);
            if (longClickDetection) {
                System.Diagnostics.Debug.WriteLine("{0}: {1} - Detected long click", DateTime.Now, name);
                OnLongClick(this);
                CancelLongClickDetection();
            }
        }

        private void CancelLongClickDetection()
        {
            System.Diagnostics.Debug.WriteLine("{0}: {1} - Canceling long click detection", DateTime.Now, name);
            LongClickDetectorHandler.RemoveCallbacks(LongClickDetector);
            longClickDetection = false;
        }

        // touch listeners
        public bool OnTouch(View v, MotionEvent e)
        {
            System.Diagnostics.Debug.WriteLine("{0}: {3} => {1} ({2})",DateTime.Now,name,UIStatus.ToString(),e.Action.ToString());
            if (!spinnerShowing)
            {
                switch (e.Action) {
                    case MotionEventActions.Move:
                        if ( ButtonState.Moving == UIStatus )
                        {
                            if (e.GetX() > swipeButtonInner.GetX() + swipeButtonInner.Width / 2 && e.GetX() + swipeButtonInner.Width / 2 < Width /*&& e.GetX() <= initialX + swipeButtonInner.Width*/)
                            {
                                swipeButtonInner.SetX(e.GetX() - swipeButtonInner.Width / 2);
                                centerText.Alpha = (1 - (float)1.3 * (swipeButtonInner.GetX() + swipeButtonInner.Width) / Width);
                                SetTrailingEffect();
                            }
                            // don't move if too close to the final position
                            if (e.GetX() + swipeButtonInner.Width / 2 > Width && swipeButtonInner.GetX() + swipeButtonInner.Width / 2 <= Width)
                            {
                                swipeButtonInner.SetX(Width - swipeButtonInner.Width);
                            }
                            // don't move if too close to the starting position
                            if (e.GetX() < swipeButtonInner.Width / 2 && swipeButtonInner.GetX() > 0)
                            {
                                swipeButtonInner.SetX(0);
                            }
                            // processing swipe event. long click does not apply here
                            CancelLongClickDetection();
                            // complete swipe if movement >90%
                            if (swipeButtonInner.GetX() + swipeButtonInner.Width > Width * 0.9)
                            {
                                EndSwipe();
                                SetActive();
                            }
                            return true;
                        } else {
                            return false;
                        }

                    case MotionEventActions.Up:
                        switch (UIStatus) {
                            case ButtonState.Expanded:
                                Reset();
                                break;
                            case ButtonState.Moving:
                                if (swipeButtonInner.GetX() + swipeButtonInner.Width > Width * 0.9)
                                {
                                    EndSwipe();
                                    SetActive();
                                }
                                else
                                {
                                    InvokeDelayedReset();
                                }
                                break;
                        }
                        // processing swipe event. long click does not apply here
                        CancelLongClickDetection();
                        return true;

                    case MotionEventActions.Down:
                        CancelDelayedReset();
                        if (e.GetX()>swipeButtonInner.GetX() && e.GetX()<swipeButtonInner.GetX()+swipeButtonInner.Width) {
                            BeginSwipe();
                            CancelLongClickDetection();
                            return true;
                        } else
                        {
                            StartLongClickDetection();
                            return true; //return false if using the built-in long click detection
                        }

                    case MotionEventActions.Cancel:
                        CancelLongClickDetection();
                        return true;

                    default:
                        InvokeDelayedReset();
                        return true;
                }
            } else {
                System.Diagnostics.Debug.WriteLine("NO EVENT HANDLING");
                return false;
            }
        }
        #endregion
    }
}